/****************************************************************/
/**@file CameraTools.cpp: 
 * Desctiption: Higher level abstraction above raw camera driver.*
 *              This code acts like driver, that provides 
 *              RT point with ball position			*
 * Dependency:							*
 ****************************************************************/
#include "hardware/OsSupp.h"

#include "images/raster.h"
#include "images/ras_comp.h"
#include "images/img_tool.h"

#include "CameraTools.h"


const int CameraTool::DEFAULT_THR = 220; ///< Default threshold value


/** Read input points from camera object */
Point *CameraTool::GetInput(int Index)
{
switch(Index)
  {
  case 0: return &VideoX;      //motor 1 position
  case 1: return &VideoY;      //motor 2 position  
  case 2: return &WrThreshold; ///< Get threshold
  default: return NULL;
  }
};


/** Periodical execution recalculates ball positions. */
int CameraTool::Execute(void)
{
int posX, posY;
 
  if(Camera==NULL) return -1;  

  try
    {
    Camera->GrabImage();		//GrabImage();

      /*int BALL_STATE = */
    if(BalSearchAlloved && BallSearch(&posX,&posY,false,false)==BALL_OK)
      {
      VideoX.SetDouble((2*posX)/(double)iWidth - 1);
      VideoY.SetDouble((-2*posY)/(double)iHeight + 1);
      }
    else
      {
      double d=0;
      MakeNaN(&d);
      VideoX.SetDouble(d);	
      VideoY.SetDouble(d);
      }
    }
  catch(...)
    {	// problem with grab
    double d=0;
    MakeNaN(&d);
    VideoX.SetDouble(d);	
    VideoY.SetDouble(d);
    return -2;
    }

return 0;
}


/** This method formats camera raw image into required format. Moreover it
  also crops image to a rectangular shape. This makes signifficant speedup of whole
  application.
  @param[in] ImageBuf Buffer that receives image data in required format.
  @param[in] mode = 0 true color image; 1 true color as gray; 2 true color as mono
                     0x11 native gray 8bit;0x12 8 bit mono  */
bool CameraTool::ReadImageAs(Raster2DAbstract *ImageBuf, int mode, bool VFlip, bool HFlip)
{
int Width, Height;
char destplanes;
int y;
int OffsetX;

  if(Camera==NULL) return false;
  if(Camera->GetBufferData()==NULL) return false;
  if(ImageBuf==NULL) return false;
  
  Width = Camera->GetWidth();
  if(Width<0) {Width=-Width;HFlip=!HFlip;}
  Height = Camera->GetHeight();
  if(Height<0) {Height=-Height;VFlip=!VFlip;}

  iWidth = Width;
  iHeight = Height;

  if(Width<=0 || Height<=0) 
	return false;  

  OffsetX = 0;

  if(mode & 0x80)
  {
    mode &= ~0x80;
    if(Width>Height)	//cut image to a rectangle
    {
      OffsetX = (Width-Height) / 2;
      iWidth = Height;
    }
  }

  destplanes = (mode<0x10) ? 24 : 8;
  if(destplanes!=ImageBuf->GetPlanes())
    return false;
  if(ImageBuf->Size1D!=iWidth || ImageBuf->Size2D!=iHeight)
  {
    ImageBuf->Allocate2D(iWidth,iHeight);
  }  
  
  // recompute the image to monochrome according to mode    

  //if(VFlip) Flip2D(ImageBuf);	// Preflip pointers it ensures same pointers order after postflip. No longer needed.

  for(y=0; y<Height; y++)
    {
    unsigned char *i, *j;
    int LoopCount = iWidth;
    int depth = Camera->GetDepth();
		//calculate buffer positions
    i = Camera->GetBufferData() + (OffsetX+y*Width)*depth/8;

    switch(depth)
      {
      case 32:
        j = (unsigned char *)ImageBuf->GetRow(Height-y-1);	//Windows defines that ARGB is vertically inverted.

        switch(mode)
          {
	  case 0x0:	//converts from BGR32 to RGB24
	    RGB32_BGR24(j,i,LoopCount);
	    break;

          case 0x1:	//converts from BGR32 to RGB gray
            BGR32_Gray24(j,i,LoopCount);
            break;

          case 0x2:	//converts from BGR32 to RGB mono
            {
            int thr3 = 3*threshold;
	    while(LoopCount-->0)
              {
	      j[2]=j[1]=j[0] = ((*i + i[1] + i[2])>=thr3) ? 255 : 0;
              j+=3;
              i+=4;
              }
            break;
            }

          case 0x11:	//converts from BGR32 to 8bit gray
            RGB32_Gray(j,i,LoopCount);
	    break;
          case 0x12:	//converts from BGR32 to 8bit mono
            {
            int thr3 = 3*threshold;
	    while(LoopCount-->0)
               {
	       *(j++) = ((*i+i[1]+i[2]) >= thr3) ? 255 : 0;
               i+=4;
               }
            break;
            }
          }
        break;

      case 24:							//RGB camera 24 bit
        j = (unsigned char *)ImageBuf->GetRow(Height-y-1);		//Windows defines that RGB is vertically inverted.
        if(j==NULL)
            break;
        
        switch(mode)
          {
	  case 0x0:	//converts from BGR to RGB
	      RGB_BGR2(j,i,LoopCount);
	      break;
  	  case 0x1:	//converts from BGR to RGB gray
              BGR_Gray24precise(j,i,LoopCount);
              break;
          case 0x2:	//converts from BGR to RGB mono
            {
            int thr3 = 3*threshold;
	    while(LoopCount-->0)
              {
	      j[2]=j[1]=j[0] = ((i[0]+i[1]+i[2])>=thr3) ? 255 : 0;
              j+=3;
              i+=3;
              }
            break;
            }
          case 0x11:	//converts from BGR to 8bit gray
	    RGB_Gray(j,i,LoopCount);
            break;
          case 0x12:	//converts from BGR to 8bit mono
            {
            int thr3 = 3*threshold;
	    while(LoopCount-->0)
               {
	         *(j++) = ((*i+i[1]+i[2]) >= thr3) ? 255 : 0;
                 i += 3;
               }
            break;
            }
          }
        break;


      case 16:	
	j = (unsigned char *)ImageBuf->GetRow(y);		// YUY2 is not flipped

        if(Camera->VideoFormat == vfRGB565)		// https://chowdera.com/2021/12/202112271902236648.html
	{
	  switch(mode)
          {
            case 0:  
	        while(LoopCount-->0)
		{
		  const uint16_t W = *(uint16_t*)i;
		  i += 2;
		  *j++ = (W>>8) & 0xF8;
		  *j++ = (W>>3) & 0xFC;
		  *j++ = (W<<3) & 0xF8;
		}
		break;
	    case 1:
	        while(LoopCount-->0)
		{
		  uint16_t W = *(uint16_t*)i;
		  i += 2;
		  j[0] = j[1] = j[2] = (((W>>8) & 0xF8) + ((W>>3) & 0xFC) + ((W<<3) & 0xF8)) / 3;
		  j+=3;
		}
		break;
	    case 2:
	        while(LoopCount-->0)
		{
		  uint16_t W = *(uint16_t*)i;
		  i += 2;
		  j[0] = j[1] = j[2] = (((W>>8) & 0xF8) + ((W>>3) & 0xFC) + ((W<<3) & 0xF8)) / 3;
		  j+=3;
		}
		break;
	    case 3:
	        while(LoopCount-->0)
		{
		  uint16_t W = *(uint16_t*)i;
		  i += 2;
		  W = (((W>>8) & 0xF8) + ((W>>3) & 0xFC) + ((W<<3) & 0xF8)) / 3;
		  j[0] = j[1] = j[2] = (W>=threshold) ? 255 : 0;
		  j+=3;
		}
		break;
	   case 0x11:
		while(LoopCount-->0)
		{
		  uint16_t W = *(uint16_t*)i;
		  i += 2;
		  *j++ = (((W>>8) & 0xF8) + ((W>>3) & 0xFC) + ((W<<3) & 0xF8)) / 3;
		}
		break;
	   case 0x12:
		while(LoopCount-->0)
		{
		  uint16_t W = *(uint16_t*)i;
		  W = (((W>>8) & 0xF8) + ((W>>3) & 0xFC) + ((W<<3) & 0xF8)) / 3;
		  i += 2;
		  *j++ = (W>=threshold) ? 255 : 0;
		}
		break;
	  }
	break;
        }

	if(Camera->VideoFormat == vfRGB555)		// https://chowdera.com/2021/12/202112271902236648.html
	{
	  switch(mode)
          {
            case 0:  
	        while(LoopCount-->0)
		{
		  const uint16_t W = *(uint16_t*)i;
		  i += 2;
		  *j++ = (W>>7) & 0xF8;
		  *j++ = (W>>2) & 0xF8;
		  *j++ = (W<<3) & 0xF8;
		}
		break;
	    case 1:
	        while(LoopCount-->0)
		{
		  uint16_t W = *(uint16_t*)i;
		  i += 2;
		  j[0] = j[1] = j[2] = (((W>>7) & 0xF8) + ((W>>2) & 0xF8) + ((W<<3) & 0xF8)) / 3;
		  j+=3;
		}
		break;
	    case 2:
	        while(LoopCount-->0)
		{
		  uint16_t W = *(uint16_t*)i;
		  i += 2;
		  j[0] = j[1] = j[2] = (((W>>7) & 0xF8) + ((W>>2) & 0xF8) + ((W<<3) & 0xF8)) / 3;
		  j+=3;
		}
		break;
	    case 3:
	        while(LoopCount-->0)
		{
		  uint16_t W = *(uint16_t*)i;
		  i += 2;
		  W = (((W>>7) & 0xF8) + ((W>>2) & 0xF8) + ((W<<3) & 0xF8)) / 3;
		  j[0] = j[1] = j[2] = (W>=threshold) ? 255 : 0;
		  j+=3;
		}
		break;
	   case 0x11:
		while(LoopCount-->0)
		{
		  uint16_t W = *(uint16_t*)i;
		  i += 2;
		  *j++ = (((W>>7) & 0xF8) + ((W>>2) & 0xF8) + ((W<<3) & 0xF8)) / 3;
		}
		break;
	   case 0x12:
		while(LoopCount-->0)
		{
		  uint16_t W = *(uint16_t*)i;
		  W = (((W>>7) & 0xF8) + ((W>>2) & 0xF8) + ((W<<3) & 0xF8)) / 3;
		  i += 2;
		  *j++ = (W>=threshold) ? 255 : 0;
		}
		break;
	  }
	break;
        }

		// YUY2 same as YUYV                
        switch(mode)
        {
          case 0:  YUYV_RGB(j, i, LoopCount);			// Converts to RGB
                   break;

          case 1:  while(LoopCount>=2)   //converts to 3xGray
                    {
                    j[0] = j[1] = j[2] = *i;
                    i += 2;
		    j += 3;
                    j[0] = j[1] = j[2] = *i;
                    i += 2;
		    j += 3;
                    LoopCount -= 2;
                    }
		  break;

          case 2:	//converts to 3x8bit mono
                  while(LoopCount>=2)   //converts to Gray
                    {
                    j[0] = j[1] = j[2] = *i>=threshold ? 255 : 0;
                    i += 2;
		    j += 3;
                    j[0] = j[1] = j[2] = *i>=threshold ? 255 : 0;
                    i += 2;
		    j += 3;
                    LoopCount -= 2;
                    }
                  break;

         case 0x11:  while(LoopCount>=2)   //converts to Gray
                    {
                    j[0] = *i;
                    i += 2;
		    j++;
                    j[0] = *i;
                    i += 2;
		    j++;
                    LoopCount -= 2;
                    }
		  break;

         case 0x12:	//converts to 8bit mono
                  while(LoopCount>=2)   //converts to Gray
                    {
                    j[0] = *i>=threshold ? 255 : 0;
                    i += 2;
		    j++;
                    j[0] = *i>=threshold ? 255 : 0;
                    i += 2;
		    j++;
                    LoopCount -= 2;
                    }
                 break;
        }
        break;


      case 12:			// I420 mode
        if(mode==0)
          {
          OffsetX &= ~1;
          YUV_RGB((unsigned char *)ImageBuf->GetRow(y),
		  Camera->GetBufferData() + (OffsetX+y*Width),	// Y canal is like 8 gray
		  Camera->GetBufferData() + (Height*Width) + (OffsetX+y*Width/2)/2,
		  Camera->GetBufferData() + 5*(Height*Width)/4 + (OffsetX+y*Width/2)/2,
		  LoopCount);	//converts to RGB
          break;
          }        
        // fallback to 8bit when mode != 0, but ImageBuf->GetRow(y) ptr is correct.
	
      case 8:	//gray camera 8 bit
        j = (unsigned char *)ImageBuf->GetRow(y);

	if(Camera->VideoFormat==vfRGB8 && Camera->pCameraPalette!=NULL)
	{
	  switch(mode)
          {
          case 0:		//converts to RGB
		while(LoopCount-->0)
		{
		  const TCameraPalette *pPal = &Camera->pCameraPalette[*i++];
		  *j++ = pPal->R;
		  *j++ = pPal->G;
		  *j++ = pPal->B;
		}
		break;
	  case 0x1:	//converts from gray to RGB gray
                while(LoopCount-->0)
		{
		  const TCameraPalette *pPal = &Camera->pCameraPalette[*i++];
		  j[2]=j[1]=j[0] = (pPal->R+pPal->G+pPal->B) / 3;
		  j += 3;		  
		}		
		break;
          case 0x2:	//converts from gray to RGB mono
	        while(LoopCount-->0)
                  {
	          const TCameraPalette *pPal = &Camera->pCameraPalette[*i++];
		  j[2]=j[1]=j[0] = (((pPal->R+pPal->G+pPal->B) / 3)>threshold) ? 255 : 0;
		  j += 3;		  
	          }
                break;
          case 0x11:	//converts from gray to 8bit gray
               while(LoopCount-->0)
		{
		  const TCameraPalette *pPal = &Camera->pCameraPalette[*i++];
		  *j++ = (pPal->R+pPal->G+pPal->B) / 3;		  
		}
               break;
          case 0x12:	//converts from RGB to 8bit mono                    
  	      while(LoopCount-->0)
                 {
		 const TCameraPalette *pPal = &Camera->pCameraPalette[*i++];
	         *(j++) = (((pPal->R+pPal->G+pPal->B) / 3)>threshold) ? 255 : 0;
                 }
              break;            
          }
          break;
	}

        switch(mode)
          {
          case 0:		//converts from gray to RGB (also gray)
	  case 0x1:	//converts from gray to RGB gray
                Conv8_24(j,i,LoopCount);
		break;
          case 0x2:	//converts from gray to RGB mono
	        while(LoopCount-->0)
                  {
	          j[2]=j[1]=j[0]= *(i++)>=threshold ? 255 : 0;
  	          j+=3;
	          }
                break;
          case 0x11:	//converts from gray to 8bit gray
               memcpy(j,i,LoopCount);	       
               break;
          case 0x12:	//converts from RGB to 8bit mono                    
  	      while(LoopCount-->0)
                 {
	         *(j++) = *(i++)>=threshold ? 255 : 0;
                 }
              break;            
          }
        break;        


      default:	     // //unsupported mode
        //wxMessageBox(_T("Unsupported image conversion!"), _T("ReadImageAs()"), wxOK| wxICON_EXCLAMATION, NULL);
        if(VFlip) Flip2D(ImageBuf); // Cleanup pointer order.
	return false;      
      }
    }

  if(VFlip) Flip2D(ImageBuf);	// postflip pointers.
  if(HFlip) Flip1D(ImageBuf);	// flip image contents.
  return true;
}


/** THIS METHOD IS A PRETTY HACK. It was dependent on a screen resolution very much.
*I released this limitation a litle bit, but I am still unsure whether this works fine.
*   @param[out] x  - x coordinate of ball (In a bitmap. No further transform!)
*   @param[out] y  - y coordinate of ball (In a bitmap. No further transform!)   */
CameraTool::BALLSTATE CameraTool::BallSearch(int* x, int* y, bool VFlip, bool HFlip)
{
  int  i, j, k, imin, jmin, imax, jmax, isize, jsize, iedge, jedge, q;
  bool around;

  if(!ReadImageAs(&IntImage,0x12,VFlip,HFlip))
    return BALL_FAIL;
  // IntImage.Height = iHeight; IntImage.Width = iWidth;  

  const int STEP    =   5;                 //< ball search step
  const int AROUND  = (60*iWidth)/480;	   //< size of expected position area (was set to 30 for resolution 255x255)  
  const int BALLMIN = (12*iWidth)/480;     //< minimum ball size (was set to 6 for resolution 255x255)
  const int BALLMAX = (75*iWidth)/480;     //< maximum ball size (was set to 30 for resolution 255x255)
  const int IGNOR   = (iWidth<=480) ? 2 : ((2*iWidth)/480); //< max. black area inside ball

  const int XMIN =  0;                     //< ball search range
  int XMAX = IntImage.Size1D;
  const int YMIN =   0;
  int YMAX = IntImage.Size2D;

  around=found;
  found=0;

  IntImage.CheckX(lastx);
  IntImage.CheckY(lasty);
  *x=lastx;                                /*  return values if not found */
  *y=lasty;

  if (around)                              /* first try around expected position */
  {
    imin = expx-AROUND;
    jmin = expy-AROUND;
    imax = expx+AROUND;
    jmax = expy+AROUND;
  }

  do
  {
    if (!around)                             /* if not around, try whole plain */
    {
      imin=XMIN;
      imax=XMAX;
      jmin=YMIN;
      jmax=YMAX;
    }

    IntImage.CheckX(imin);		/* sanity check */
    IntImage.CheckX(imax);
    IntImage.CheckY(jmin);
    IntImage.CheckY(jmax);

    for (j=jmin; j<jmax && !found; j+=STEP)
    {
      Raster1DAbstract *R1Row = IntImage.GetRowRaster(j);
      if(R1Row==NULL) continue;

      for (i=imin; i<imax && !found; i+=STEP)      
      {							/* find first white point */
        if(!(found=(R1Row->GetValue1D(i)>0))) continue;
                                                         /* get ball size & centre */
        for (k=0,q=j; (k<IGNOR) && (q>YMIN); q--)        /* y <- */
        {
          if(IntImage.iswhite(i,q)) k=0;
          else
          {
            if(IntImage.iswhite(i+1,q)) continue;
            if(IntImage.iswhite(i-1,q)) continue;
            k++;
          }
        }
        jedge = q + IGNOR;

        for (k=0,q=j; (k<IGNOR) && (q<YMAX); q++)        /* y -> */
        {
          if(IntImage.iswhite(i,q)) k = 0;
          else
          {
            if(IntImage.iswhite(i+1,q)) continue;
            if(IntImage.iswhite(i-1,q)) continue;          
            k++;
          }
        }
        jsize = (q-IGNOR)-jedge+1;

        if (jsize>BALLMAX)                               /* if too big exit */
        {
          found=0;
          return(BALL_BIG);
        }

        for (k=0,q=i; (k<IGNOR) && (q>XMIN); q--)        /* x <- */
        {
          if(R1Row->GetValue1D(q)>0) k=0;
          else
          {
            if(IntImage.iswhite(q,j+1)) continue;
            if(IntImage.iswhite(q,j-1)) continue;
            k++;
          }
        }
        iedge = q+IGNOR;

        for(k=0,q=i; (k<IGNOR) && (q<XMAX); q++)        /* x -> */
        {
          if(R1Row->GetValue1D(q)>0) k=0;
          else
          {
            if(IntImage.iswhite(q,j+1)) continue;
            if(IntImage.iswhite(q,j-1)) continue;
            k++;
          }
        }
        isize = (q-IGNOR)-iedge+1;

        if (isize>BALLMAX)                               /* if too big exit */
        {
          found=0;
          return(BALL_BIG);
        }        

        if (isize<BALLMIN || jsize<BALLMIN) found=0;     /* if too small try again */
      }
    }
    around=!around;
  }
  while(!around);

  if (!found) return(BALL_NONE);

  i = iedge+isize/2;                                       /* compute ball centre to [i,j] */
  j = jedge+jsize/2;

  expx = i+(i-lastx);
  if(expx>XMAX-AROUND) expx=XMAX-AROUND;                /* compute expected position */
  if(expx<XMIN+AROUND) expx=XMIN+AROUND;

  expy = j+(j-lasty);
  if(expy>YMAX-AROUND) expy=YMAX-AROUND;
  if(expy<YMIN+AROUND) expy=YMIN+AROUND;

  lastx=i;
  lasty=j;

  *x=i;                             /* return values */
  *y=j;
  return(BALL_OK);
}
